/* $Id: linfit.c,v 1.18 1998/09/19 04:10:07 jt Exp $ */
/* linfit.c */
/* make a linear fit to a set of data
 * input is n rows of m+1 columns, like this
 * y[1]  x[1][1] x[2][1] ... x[m][1]
 * y[2]  x[1][2] x[2][2] ... x[m][2]
 * ...
 * y[n]  x[1][n] x[2][n] ... x[m][n]
 *
 * output is a vector a[1..m] which
 * is the best fit to:
 * y = a[1]x[1] + a[2]x[2] + ... a[m]x[m]
 *
 * alternative out can be predicted y[], or residuals 
 *
 * 5/26/94 jt@lanl.gov
 * 4/ 2/95 various mods
 * 7/21/95 insample/outsample
 */
/* Matrix manipulation is always tricky, and I have adopted a few conventions
 * to help keep things from getting too out of hand:
 * Variable names
 *    n     number of data points
 *    i     loop variable: always used for looping over data points
 *    j,k   loop variables: for looping over matrix indices
 *    m     number of free input parameters (size of argument vector X)
 */

#include <stdio.h>
#include <math.h>
#include "opt.h"
#include "nrutil.h"


/* prototypes */

void
sqmatsolve(float **a,
	   float *b,
	   float *x,
	   int m);
int
svd_matsolve(float **x, float *y, float *a, int n, int m);

static int
norm_matsolve(float **x, float *y, float *a, int n, int m);

float
msq_error( float **x, 
	   float *y,
	   float yo,
	   float *a,
	   int n,
	   int m );

float
variance( float *y, int n);

FILE *
openFile(char *s,
	 char *rw,
	 FILE *fpDefault);

void 
ignore_eol(FILE *fp);

/* globals */

int n=1000,mm=1,M=0;
char *fname=NULL;
char outf = 'a';
int constf = 1;
int subtract_means = 1;
int verbose=0;
int Yarg=0;
double x_sigma=0;
char *Xlist=NULL;
char *Flist=NULL;
char *Vlist=NULL;
int use_normal_eqs=1;

int run();
main(int argc,
     char **argv)
{
    optreg(&n,INT,'n',"number of data points (rows) (upper bound)");
    optreg(&mm,INT,'m',"number of arguments to linear fit");
    optreg(&Yarg,INT,'Y',"Y column (Y=0 --> first column)");
    optreg(&Xlist,VSTRING,'X',
                "list of columns for X arguments, eg 1-3,5,8");
    optreg(&Flist,VSTRING,'F',
                "list of rows for fitting, eg 1-100");
    optreg(&Vlist,VSTRING,'V',
                "list of rows for verification, eg 101-200");
    optreg(&x_sigma,DOUBLE,'z',
                "expected noise in X arguments");
    optreg(&M,INT,'M',"total number of args in data set (columns-1) (?M help)");
    opthelp(&M,
	    "note M=0 is okay, but if M+1 > cols in data, then silent disaster!");
    optreg(&fname,VSTRING,'I',"input file name");
    optreg(&fname,VSTRING,'\0',"input file name");
    optreg(&outf,CHAR,'o',"output type (?o for help)");
    opthelp(&outf,
	    "(a) solution vector, (e/E)rr, (r)esiduals, (y) predicted out");
    optreg(&constf,FLAG,'c',"include constant term");
    optreg(&subtract_means,FLAG,'s',"subtract means");
    optreg(&use_normal_eqs,FLAG,'N',"use normal equations");
    optreg(&verbose,INTLEVEL,'v',"verbose");
    optRun(run);
    opt(&argc,&argv);
    run();
}
int run()
{
    int i,j,k;
    int m;
    int nfit;			/* number of rows in fitting matrix */
    int nval;			/* number of rows in validation matrix */
    float ypred,err,var;
    FILE *fp;			/* input file */
    float **x, **xfit, **xval;	/* design matrix */
    double *zz;                 /* a row of data */
    float **z;                  /* full data matrix */
    float *y,*yfit,*yval;	/* y vector */
    float *a;			/* solution vector */
    float **xtx;		/* transpose(design) * design */
    float *xty;			/* transpose(design) * y      */
    float yo,*xo;               /* ymean, xmean */
    int *kvec;

    /* m is size of argument vector 
     *
     * mm is number of columns in design matrix (includes an extra
     * column of just 1's if constf is true
     *
     * M is number of argument columns in the actual input data set;
     * M+1 is total number of columns in the input data set.
     * Actually, there can be more than M+1 columns in the input data sets,
     * but all subsequent columns will be ignored.  However, beware:
     * if M+1 is larger than the number of columns in the data file,
     * then you'll get the wrong answer ***** with no error message *****
     * 
     */

    constf = (constf ? 1 : 0);
    m = mm + constf;
    M = (M == 0 ? mm : M);

    y = vector(1,n);
    x = matrix(1,n,1,m);
    zz = dvector(0,M);
    if (outf == 'r' || outf == 'y')
	z = matrix(1,n,0,M);
    a = vector(1,m);

    /* set up kvec: kvec[j] = k */
    /* k is actual column number */
    /* j is the arg number */
    kvec = ivector(1+constf,m);
    if (Xlist != NULL) {
	j = inlist2array(Xlist,m-constf,kvec+1+constf);
	if (j < m-constf) {
	    fprintf(stderr,"Invalid specification of -X, the list \"%s\"",Xlist);
	    fprintf(stderr," should contain %d values\n",m-constf);
	    exit(2);
	}    
	for (j=1+constf; j<=m; ++j) {
	    if (kvec[j] > M) {
		fprintf(stderr,"Invalid specification of -X,");
		fprintf(stderr," the list \"%s\"",Xlist);
		fprintf(stderr," contains too large a value: %d \n",kvec[j]);
		fprintf(stderr,"Column %d set to zero\n",j);
		kvec[j]=0;
	    }
	}
    } else {
	if (M < mm) {
	    fprintf(stderr,"Cannot specify M (%d) smaller than m (%d)\n",M,mm);
	    exit(4);
	}
	for (j=1+constf; j<=m; ++j) {
	    kvec[j] = j-constf;
	}
    }


    if (verbose>1) {
	for (j=1+constf; j<=m; ++j)
	    fprintf(stderr,"kv[%d]=%d%s",j,kvec[j],(j==m ? "\n" : ", "));
    }

    /* open the data file */
    fp = openFile(fname,"r",stdin);

    /* read the data from the file */
    for (i=1; i<=n; ++i) {
	/* read the first 1+M columns */
	for (k=0; k<=M; ++k) {
	    if (1 != fscanf(fp,"%lf",&zz[k]))
		break;
	}
	if (k <= M) 
	    break;
	/* ignore subsequent columns */
	ignore_eol(fp);

	/* take the y value as the Yarg'th (0th) column */
	y[i] = zz[Yarg];

	/* read each of the x[1..m][i] matrix elements */
	/* first do j=1 */
	if (constf)
	    x[i][1] = 1.0;
	    
	for (j = 1 + constf; j<=m; ++j) {
	    x[i][j] = zz[ kvec[j] ];
	}

	if (verbose > 2) {
	    fprintf(stderr,"%g ",y[i]);
	    for (j=1; j<=m; ++j) {
		fprintf(stderr,"%g ",x[i][j]);
	    }
	    fprintf(stderr,"\n");
	}
	if (outf == 'r' || outf == 'y') {
	    /* save the actual columnar data */
	    for (k=0; k<=M; ++k) {
		z[i][k] = zz[k];
	    }
	}
    }
    n = i-1;
    if (verbose)
	fprintf(stderr,"n=%d, M=%d\n",n,M);
    if (n==0) {
	fprintf(stderr,"Error, no input data\n");
	exit(3);
    }
    /* done with the input row vector */
    free_dvector(zz,0,M);

    /* ok, you can close the input file now */
    fclose(fp);

    /* Now make the fitting matrix and vector: xfit, yfit. */
    if (Flist != NULL) {
	/* nfitvec is the vector of rows used in fitting; thus
	 * nfitvec[ifit] = i
	 */
	int *nfitvec,ifit;
	nfitvec = ivector(1,n);
	nfit = inlist2array(Flist,n,nfitvec+1);
	yfit = vector(1,nfit);
	xfit = matrix(1,nfit,1,m);
	for (ifit=1; ifit<=nfit; ++ifit) {
	    i = nfitvec[ifit];
	    yfit[ifit] = y[i];
	    for (j=1; j<=m; ++j) {
		xfit[ifit][j] = x[i][j];
	    }
	}
	free_ivector(nfitvec,1,n);
    } else { /* if Flist == NULL */
	yfit = y;
	xfit = x;
	nfit = n;
    }

    /* Now make the validation matrix and vector: xval, yval. */
    if (Vlist != NULL) {
	/* nvalvec is the vector of rows used in validating; thus
	 * nvalvec[ival] = i
	 */
	int *nvalvec,ival;
	nvalvec = ivector(1,n);
	nval = inlist2array(Vlist,n,nvalvec+1);
	yval = vector(1,nval);
	xval = matrix(1,nval,1,m);
	for (ival=1; ival<=nval; ++ival) {
	    i = nvalvec[ival];
	    yval[ival] = y[i];
	    for (j=1; j<=m; ++j) {
		xval[ival][j] = x[i][j];
	    }
	}
	free_ivector(nvalvec,1,n);
    } else { /* if Vlist == NULL */
	yval = y;
	xval = x;
	nval = n;
    }

    /* subract out the means */
    yo = 0;
    xo = vector(1,m);
    if (subtract_means) {
	for (i=1; i<=nfit; ++i) {	
	    yo += yfit[i];
	}
	yo /= n;
	for (i=1; i<=nfit; ++i) {	
	    yfit[i] -= yo;
	}
	xo[1]=0;
	for (j=1 + constf; j<=m; ++j) {			 
	    xo[j] = 0;
	    for (i=1; i<=nfit; ++i) {
		xo[j] += xfit[i][j];
	    }
	    xo[j] /= n;
	    for (i=1; i<=nfit; ++i) {
		xfit[i][j] -= xo[j];
	    }
	}

	if (verbose) {
	    fprintf(stderr,"MEANS: Y = %g, X = [ ",yo);
	    for (j=1+constf; j<=m; ++j)
		fprintf(stderr,"%g ",xo[j]);
	    fprintf(stderr,"]\n");
	}
	
    }

    if (use_normal_eqs) {
	norm_matsolve(xfit,yfit,a,nfit,m);
    } else {
	/* if not normal matrix method, use straight SVD on the data */
	svd_matsolve(xfit,yfit,a,nfit,m);
    }
    
    /* now add back the means */
    if (subtract_means) {
	for (i=1; i<=nfit; ++i) {	
	    yfit[i] += yo;
	}
	for (j=1 + constf; j<=m; ++j) {			 
	    for (i=1; i<=nfit; ++i) {
		xfit[i][j] += xo[j];
	    }
	}
	/* and fix the value of the constant term: a[0] */
	for (j=1 + constf; j<=m; ++j) {			 
	    yo -= a[j]*xo[j];
	}
	if (constf) {
	    a[1] += yo;
	    yo = 0;
	}
	free_vector(xo,1,m);
    }
	
    if (verbose) {
	if (subtract_means && !constf)
	    fprintf(stderr,"yo=%g ",yo);
	for (j=1; j<=m; ++j)
	    fprintf(stderr,"a[%d]=%g%s",j-constf, a[j],
		    (j==m ? "\n" : ", "));
	/* first report the error in the whole data set */
	err = msq_error(x,y,yo,a,n,m);
	fprintf(stderr,"Full rms = %g\n",sqrt(err));
	err = msq_error(xfit,yfit,yo,a,nfit,m);
	fprintf(stderr,"Fit  rms = %g\n",sqrt(err));
	err = msq_error(xval,yval,yo,a,nval,m);
	fprintf(stderr,"Test rms = %g\n",sqrt(err));
	if (x_sigma > 0) {
	    float zerr;
	    zerr = err;
	    for (j=1+constf; j<=m; ++j) {
		zerr += a[j]*a[j]*x_sigma*x_sigma;
	    }
	    fprintf(stderr,"zrms = %g\n",sqrt(zerr));
	}
	
    }

    /* output */
    switch (outf) {

    case 'e':
	/* report error in validation set */
	err = msq_error(xval,yval,yo,a,nval,m);
	fprintf(stdout,"%g\n",sqrt(err));
	break;

    case 'E':
	/* report error in validation set */
	/* normalized by variance of validation set */
	err = msq_error(xval,yval,yo,a,nval,m);
	var = variance(yval,nval);
	/* fprintf(stderr,"nval=%d, var=%g\n",nval,var); */
	fprintf(stdout,"%g\n",sqrt(err/var));
	break;

    case 'y':
	for (i=1; i<=n; ++i) {
	    ypred = yo;
	    for (j=1; j<=m; ++j) {
		ypred += a[j]*x[i][j];
	    }
	    fprintf(stdout,"%g ",ypred);
	    for (k=1; k<=M; ++k)
		fprintf(stdout,"%g ",z[i][k]);
	    fprintf(stdout,"\n");
	}
	break;

    case 'r':
	for (i=1; i<=n; ++i) {
	    ypred = yo;
	    for (j=1; j<=m; ++j) {
		ypred += a[j]*x[i][j];
	    }
	    fprintf(stdout,"%g ",y[i]-ypred);
	    for (k=1; k<=M; ++k)
		fprintf(stdout,"%g ",z[i][k]);
	    fprintf(stdout,"\n");
	}
	break;

    case 'A':
	for (i=1+constf; i<=m; ++i)
	    fprintf(stdout,"%g\n",a[i]);
	break;
    default:
    case 'a':
	for (i=1; i<=m; ++i)
	    fprintf(stdout,"%g\n",a[i]);
	break;
    }

    free_vector(a,1,m);
    free_vector(y,1,n);
    free_matrix(x,1,n,1,m);
    if (outf == 'r' || outf == 'y')
	free_matrix(z,0,M,1,n);
    if (Flist != NULL) {
	free_matrix(xfit,1,nfit,1,m);
	free_vector(yfit,1,nfit);
    }
    if (Vlist != NULL) {
	free_matrix(xval,1,nval,1,m);
	free_vector(yval,1,nval);
    }
    return 0;
}

void 
ignore_eol(FILE *fp)
{
    int c;
    while( (c = getc(fp)) != '\n' && c != EOF)
	/* nul */;
}
static int
norm_matsolve(float **x, float *y, float *a, int n, int m)
{
    int i,j,k;
    float **xtx, *xty;

    /* get xtx and xty */
	
    xtx = matrix(1,m,1,m);
    xty = vector(1,m);
	
    for (k=1; k<=m; ++k) {
	for (j=1; j<=m; ++j) {
	    xtx[k][j] = 0;
	    for (i=1; i<=n; ++i) {
		xtx[k][j] += x[i][k]*x[i][j];
	    }
	}
    }
	
    for (j=1; j<=m; ++j) {
	xty[j] = 0;
	for (i=1; i<=n; ++i) {
	    xty[j] += x[i][j]*y[i];
	}
    }

#define ADJUST_NORMAL_BY_ZERR
#ifdef ADJUST_NORMAL_BY_ZERR
    /* adjust normal equations according to noise level expected in X */
    for (j=1+constf; j<=m; ++j)
	xtx[j][j] += x_sigma*x_sigma;
#endif
	
    /* solve (xtx)*a = (xty); solution is returned in a */
	
    sqmatsolve(xtx,xty,a,m);
	
    free_matrix(xtx,1,m,1,m);
    free_vector(xty,1,m);
}

float
variance( float *y, int n)
{
    int i;
    float mean,var;
    mean = 0.0;
    for (i=1; i<=n; ++i) {
	mean += y[i];
    }
    mean /= n;
    var = 0.0;
    for (i=1; i<=n; ++i) {
	var += (y[i]-mean)*(y[i]-mean);
    }
    var /= n;
    return var;
}
    
float
msq_error( float **x, 
	   float *y,
	   float yo,
	   float *a,
	   int n,
	   int m )
{
    int i,j;
    float ypred;
    float err = 0.0;
    for (i=1; i<=n; ++i) {
	ypred = yo;
	for (j=1; j<=m; ++j) {
	    ypred += a[j]*x[i][j];
	}
	err += (ypred-y[i])*(ypred-y[i]);
    }
    err /= n;
    return err;
}

