/* dlnfact.c */
/*
    Copyright (C) 1997 James Theiler, jt@lanl.gov

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program (it is in the file COPYING); if not, write 
    to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, 
    MA 02139, USA.
                       
*/
#include <stdio.h>
#include <math.h>
#include "dlnfact.h"

#ifndef DFACT_TABSIZE
#define DFACT_TABSIZE 500
#endif

static int    dlnfact_nmax=-1;
static double dlnfact_table[DFACT_TABSIZE];

double dlnfact(int n)
{
    int i;
    double df;

    if (n < 0) return 1.0;

    if (n <= dlnfact_nmax) {
        /* table is okay */
        return dlnfact_table[n];
    }
    /* Get the table started */
    if (dlnfact_nmax < 0) {
        dlnfact_table[0]=0.0;
        dlnfact_nmax=0;
    }
    if (n < DFACT_TABSIZE) {
        /* need to update the table */
        for (i=dlnfact_nmax+1; i<=n; ++i)
            dlnfact_table[i] = dlnfact_table[i-1] + log((double)i);
        dlnfact_nmax=n;
        return dlnfact_table[n];
    } else {
        /* too big for the table...use Stirlings approximation */
        return dlnstirling(n);
    }

}
double dlnbinom(int t, int b)
{
    if (b>t || b<0 || t<0) return -1;
    return dlnfact(t) - dlnfact(b) - dlnfact(t-b);
}
    
double dbinom(int t, int b)
{
    /* compute the binomial coeff with t on top and b on bottom */
    int i;
    double binom;
    
    if (b>t || b<0 || t<0) return 0;

    return floor(exp(dlnfact(t) - dlnfact(b) - dlnfact(t-b))+0.5);
    
#if 0 /* This method is now depricated... */    
    /* Choose b to be the smaller of b and t-b */
    b = b < t-b ? b : t-b;
    
    if (b<DFACT_TABSIZE) {
        binom = floor(exp(-dlnfact(b))+0.5)
        for (i=t-b+1; i<=t; ++i)
            binom *= (double)i;
    } else {
        binom = 1.0;
        for (i=t-b+1; i<=t; ++i)
            binom *= (double)i;
        for (i=1; i<=b; ++i)
            binom /= (double)i;
        fprintf(stderr,"Difficult to trust this value: Binom(%d %d)=%g\n",
                t,b,binom);
    }
    return binom;
#endif    
}

#define LNTWOPIBYTWO 0.91893853320467274177
double dlnstirling(int n)
{
    /* return log(n!) using Stirling's approximation */
    /* Note, even for n=10, this is accurate to six decimal places */
    double z;
    z = n+1.0;
    if (z < 1.0) return 1.0;
    
    return (z-0.5)*log(z) - z + LNTWOPIBYTWO + 1.0/(12.0*z);
}
        

#ifdef MAIN
main(int argc, char **argv)
{
    int n,m;
    double d;
    switch (argc) {
    case 2:
        n = atoi(argv[1]);
        d = dlnfact(n);
        fprintf(stdout,"log(%d!) = %.20g\n",n,d);
        fprintf(stdout,"Stirling = %.20g\n",  dlnstirling(n));
        fprintf(stdout,"Diff     = %g\n",d-dlnstirling(n));
        if (d < 706.6) 
            fprintf(stdout,"    %d!  = %g\n",n,exp(d));
        break;
    case 3:
        n = atoi(argv[1]);
        m = atoi(argv[2]);
        fprintf(stdout,"Binomial(%d,%d) = exp(%g) = %g\n",n,m,dlnbinom(n,m),
                exp(dlnbinom(n,m)));
        fprintf(stdout,"Binomial(%d,%d) = %g\n",n,m,dbinom(n,m));
        break;
    case 1:
    default:
        fprintf(stderr,"Usage: %s n [m]\n",argv[0]);
        fprintf(stderr,"       %s n    --> n!\n",argv[0]);
        fprintf(stderr,"       %s n m  --> n!/(m!(n-m)!)\n",argv[0]);
        break;
    }
    return 0;
}
#endif /* MAIN */
        
    
