/* like.c */
/* log likelihood for a single point */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "like.h"
#include "linmin.h"
#include "dlnfact.h"

double qromb(double (*func)(double), double, double);

#define GAUSSIAN_TIME_ERROR 0
#define USE_ROMBERG 1


extern double STEPS_PER_TAU;
const double MIN_LOG_LIKE=-200;
const double IRANGE_IN_TAUS=10;
const double BAILOUT_FRACTION=0.001;

double sinh(double x)
{
    return (exp(x)-exp(-x))/2;
}
#if GAUSSIAN_TIME_ERROR
double
time_like(double t, double tau)
{
    return exp(-t*t/(2*tau*tau))/tau;
}
double
pois_loglike(double mu, int n)
{
    return -(n-mu)*(n-mu)/(400);
}
#else    
double
time_like(double t, double tau)
{
    /* Likelihood for data at time -t,
     * given model at time zero
     */
    /* fprintf(stderr,"time_like: t/tau=%g/%g=%g\n",t,tau,t/tau); */
    if (t<-0.5) return 0;
    if (tau == 0) {
        if (t<0.5) return 1.0;
        else       return 0.0;
    }
    if (t > tau*20) return 0;
    if (t< 0.5) return 1.0-exp(-(t+0.5)/tau);
    return exp(-t/tau)*sinh(0.5/tau)*2.0;
}
double
pois_loglike(double mu, int n)
{
    if (mu<0 && n==0) return 0;
    if (mu<0) return MIN_LOG_LIKE;
    return n*log(mu) - mu - dlnfact(n);
}
#endif 
double
pois_like(double mu, int n)
{
    double ll=0;
    ll = pois_loglike(mu,n);
    if (ll<=MIN_LOG_LIKE) return 0;
    return exp(ll);
}

/* Assumes data value is (to,n) */

#if USE_ROMBERG
double
full_like(double a, double b, double tau, double to, int n)
{
    double tlo,thi;
    flike_init(a,b,tau,to,n);
    tlo=(n-4*sqrt(1+(double)n)-b)/a;
    thi=(n+4*sqrt(1+(double)n)-b)/a;
    if (thi > to+0.5)
        thi = to+0.5;
    if (tlo < to-IRANGE_IN_TAUS*tau-1)
        tlo = to-IRANGE_IN_TAUS*tau-1;
    
    /* fprintf(stderr,"qromb BEGIN: to=%g,n=%d (%g,%g) [%g,%g]\n",
            to,n,a,b,tlo,thi);
    */
    return qromb(flike,tlo,thi);
}
#else
double
full_like(double a, double b, double tau, double to, int n)
{
    double t,mu,l,fl,lmax;
    double pl,tl;
    double dt;
    
    dt = (tau+1)/STEPS_PER_TAU;
    fl = 0;
    lmax=0;
    for (t=0; t>to-IRANGE_IN_TAUS*tau-1; t -= dt) {
        mu = a*t + b;
        pl = pois_like(mu,n);
        tl = time_like(to-t,tau);
        l = pl*tl;
        fl += l;
        if (l < BAILOUT_FRACTION*lmax) break;
        if (l > lmax) { lmax=l; }
    }
    lmax=0;
    for (t=dt; t<to+IRANGE_IN_TAUS*tau+1; t += dt) {
        mu = a*t + b;
        pl = pois_like(mu,n);
        tl = time_like(to-t,tau);
        l = pl*tl;
        fl += l;
        if (l < BAILOUT_FRACTION*lmax) break;
        if (l > lmax) { lmax=l; }
    }
    return dt*fl;
}
#endif
double
full_loglike_array(double a, double b, double tau, int N, Data_t *data)
{
    int i;
    double like,loglike=0;
    for (i=0; i<N; ++i) {
        like = full_like(a,b,tau,data[i].t,data[i].n);
        if (like <= 0) {
            like=-1000;
            /* fprintf(stderr,"Likelihood of 0 noted...and dealt with\n"); */
        } else {
            like = log(like);
        }
        loglike += like;
    }
    return loglike;
}



