#include <math.h>

#define USE_TRIG_FOR_CIRCLE    0 /* it's faster on Pentium, not on Sun */
#define USE_VON_NEUMANN_TRICK  1 /* avoids a sqrt() */

/* Get a random point on the unit sphere */
#ifndef PI
#define PI 3.1415926535897932
#endif

void rand_unitcircle(double *px, double *py)
{
#if USE_TRIG_FOR_CIRCLE
    /* This is the obvious solution... */
    /* It ain't clever, but since sin/cos are often hardware accelerated,
     * it can be faster than von Neumann's solution below; it is faster
     * on my Pentium at home, and slower on my Sun Sparc 20 at work
     */     
    double t;
    t = 2.0*PI*rand_unif();
    *px = cos(t);
    *py = sin(t);
#else /* dont USE_TRIG_FOR_CIRCLE */
    /* Avoids trig.  But it does take an average of 8/pi = 2.55 calls
     * to rand_unif() instead of just one, as above.
     */
    double u,v,s;
    do {
        u = -1 + 2*rand_unif();
        v = -1 + 2*rand_unif();
        s = u*u+v*v;
    } while (s > 1.0 || s == 0.0);
#if USE_VON_NEUMANN_TRICK 
    /* Heres a solution that goes back to von Neumann.  See Knuth, v2,
     * 3rd ed, p140 (exercise 23).  Note, no sin, cos, or sqrt !
     */
    *px = (u*u-v*v)/s;
    *py = 2*u*v/s;
#else  /* don't USE_VON_NEUMANN_TRICK */  
    /* Here is the more straightforward approach */
    /* Fewer total operations, but one of them is a sqrt */
    s = sqrt(s);
    *px = u/s;
    *py = v/s;
#endif  /* end USE_VON_NEUMANN_TRICK */
#endif  /* end USE_TRIG_FOR_CIRCLE */
}

void rand_unitsphere_cap(double *px, double *py, double *pz, double h)
{
    double s,a;
    /* This is a variant of the algorithm for computing a random point
     * on the unit sphere; the algorithm is suggested in Knuth, v2,
     * 3rd ed, p136; and attributed to Robert E Knop, CACM, 13 (1970),
     * 326.
     */
    
    /* Begin with the polar method for getting x,y inside a unit circle
     */
    do {
        *px = -1 + 2*rand_unif();
        *py = -1 + 2*rand_unif();
        s = (*px)*(*px)+(*py)*(*py);
    } while (s > 1.0 || s == 0.0);

    /* For the full unit sphere, one would now take z=-1+2*s, but instead
     * we want z distributed between 1-h and 1.  Since s is uniformly
     * distributed between 0 and 1, we just rescale */
    
    *pz = 1-h + s*h;

    /* rescale x,y so that they are on circle of radius sqrt(1-z*z),
     * currently their radius is sqrt(s), so rescaling factor is
     * sqrt((1-z*z)/s)
     */
    s = sqrt((1.0-(*pz)*(*pz))/s);
    *px *= s; 
    *py *= s;
    /* and voila! */
#if DEBUG
    fprintf(stderr,"rand_unitsphere_cap: r^2=%g\n",
            (*px)*(*px)+(*py)*(*py)+(*pz)*(*pz));
#endif
}
    
void rand_unitsphere(double *px, double *py, double *pz)
{
    rand_unitsphere_cap(px,py,pz,2.0);
}


